from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.moviepy.video_effect import MoviepySetPosition
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionSetPosition
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_corner import CoordinateCorner
from yta_general_utils.math.rate_functions import RateFunction
from yta_multimedia.video.frames import VideoFrameExtractor
from yta_general_utils.programming.parameter_validator import NumberValidator
from yta_general_utils.programming.error_message import ErrorMessage
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy.editor import ImageClip, CompositeVideoClip, concatenate_videoclips


class TransitionType(Enum):
    """
    Class to represent the transition types (the way we
    join the videos between a transition).
    """
    FREEZE = 'freeze'
    """
    Freeze the last frame of the first video and the first
    frame ot he second video and make the transition with
    those static frames so the duration is extended by the
    transition duration.
    """
    PLAYING = 'playing'
    """
    Keep both video playing while the transition is run.
    """

class Transition:
    """
    Class to encapsulate all the transition methods we
    handle in our system. Each transition is a way of
    connecting two different videos.

    This class is built to be used within the
    TransitionGenerator class as parameter to build 
    videos with those transitions.
    """
    @staticmethod
    def handle_transition(video1, video2, duration: float = 0.5, transition_type: TransitionType = TransitionType.FREEZE, transition_method: type = 'Transition.slide'):
        """
        Create the transition between 'video1' and 'video2' and 
        return the items in the next order: video1, video2,
        transition.
        """
        video1 = VideoParser.to_moviepy(video1)
        video2 = VideoParser.to_moviepy(video2)

        if not NumberValidator.is_positive_number(duration):
            raise Exception(ErrorMessage.parameter_is_not_positive_number('duration'))
        
        if not PythonValidator.is_class_staticmethod(Transition, transition_method):
            # TODO: We have some methods we don't accept
            raise Exception('The provided "transition_method" parameter is not an static method of the Transition class.')
        
        transition = None
        # Handle transition and videos
        if transition_type == TransitionType.FREEZE:
            # Original videos are not modified
            transition_first_clip = ImageClip(VideoFrameExtractor.get_last_frame(video1), duration = duration).set_fps(60)
            transition_second_clip = ImageClip(VideoFrameExtractor.get_first_frame(video2), duration = duration).set_fps(60)
        elif transition_type == TransitionType.PLAYING:
            if duration > (video1.duration / 2) or duration > (video2.duration / 2):
                # TODO: Make this Exception more description (which video is wrong)
                # and talk about that we use the half of the provided duration
                raise Exception(f'The provided "duration" parameter {str(duration)} is not valid according to the provided video duration.')
            
            half_duration = duration / 2
            transition_first_clip = video1.subclip(video1.duration - half_duration, video1.duration)
            transition_second_clip = video2.subclip(video2.duration - half_duration, video2.duration)
            video1 = video1.subclip(0, video1.duration - half_duration)
            video2 = video2.subclip(0, video2.duration - half_duration)
        
        transition = transition_method(transition_first_clip, transition_second_clip, duration)

        return video1, video2, transition

    @staticmethod
    def slide(video1, video2, duration: float = 0.5):
        """
        Simple transition in which the last frame of the provided 'video1'
        is replaced by the first frame of the provided 'video2' by sliding
        from right to left.
        """
        video1 = VideoParser.to_moviepy(video1)
        video2 = VideoParser.to_moviepy(video2)

        if not NumberValidator.is_positive_number(duration):
            raise Exception(ErrorMessage.parameter_is_not_positive_number('duration'))
        # TODO: Maybe put some limit (10s or similar) (?)

        # TODO: I was working before in setting the position
        # constants so I could use them to easily handle slides
        # from different sides of the screen

        # Transition from last frame of video1 to first of video2
        transition_clip_1 = MoviepySetPosition(CoordinateCorner(0, 0), CoordinateCorner(-video1.w, 0), TFunctionSetPosition.linear, RateFunction.linear).apply(video1)
        # TODO: Maybe set 'transition_clip_1.pos' instead of (0, 0)
        transition_clip_2 = MoviepySetPosition(CoordinateCorner(video2.w, 0), CoordinateCorner(0, 0), TFunctionSetPosition.linear, RateFunction.linear).apply(video2)

        return CompositeVideoClip([
            transition_clip_1,
            transition_clip_2
        ])

class TransitionGenerator:
    """
    Class to simplify and encapsulate the functionality
    related to transition between videos
    """
    # TODO: This method is no longer being used due to the
    # code format we need to use due to we could change the
    # original videos so we need to update those instances
    @staticmethod
    def get_transition_clips(videos, duration: float = 0.5, transition_type: TransitionType = TransitionType.FREEZE, transition_method: type = Transition.slide):
        """
        Generate all the transition clips between the provided
        'videos' and with the also provided 'transition_method'
        (that must be a transition method within the Transition
        class).
        """
        # TODO: This method is a candidate for internal use only
        if not NumberValidator.is_positive_number(duration):
            raise Exception(ErrorMessage.parameter_is_not_positive_number('duration'))
        
        if not PythonValidator.is_class_staticmethod(Transition, transition_method):
            raise Exception('The provided "transition_method" parameter is not an static method of the Transition class.')
        
        transition_type = TransitionType.to_enum(transition_type)
        videos = [VideoParser.to_moviepy(video) for video in videos]

        transitions = []

        for video1, video2 in zip(videos, videos[1:]):
            transitions.append(TransitionGenerator.get_transition_clip(video1, video2, duration, transition_method, transition_type))

        return transitions

    @staticmethod
    def apply(videos, duration: float = 0.5, transition_type: TransitionType = TransitionType.FREEZE, transition_method: type = 'Transition.slide'):
        """
        Build a transition by using the provided 'transition_method' 
        (that must be a transition method within the Transition class)
        and put it between all the provided videos that are played
        completely before and after the transition.

        So, the sequence is the next.
        1. The 'videoX' is played completely.
        2. The transition is played completely.
        3. Go to step 1 with the next video until last one
        4. Last video is played completely
        """
        # TODO: Make the 'transition_method' more dynamic to
        # accept different transitions between clips so we
        # can handle all of them separately
        # TODO: 'transitions' array should be len(videos) - 1
        # and all members should be accepted transition methods
        if not NumberValidator.is_positive_number(duration):
            raise Exception(ErrorMessage.parameter_is_not_positive_number('duration'))

        if not PythonValidator.is_class_staticmethod(Transition, transition_method):
            raise Exception('The provided "transition_method" parameter is not an static method of the Transition class.')

        videos = [VideoParser.to_moviepy(video) for video in videos]
        transition_type = TransitionType.to_enum(transition_type)

        clips_to_concat = []
        prev_video1 = None
        for video1, video2 in zip(videos, videos[1:]):
            # TODO: The video2 changes, but I don't know if the for
            # loop keeps the object changed in the next iteration.
            # If it does just remove the 'prev_video1' var
            if prev_video1 is not None:
                video1 = prev_video1

            video1, video2, transition = Transition.handle_transition(video1, video2, duration, transition_type, transition_method)
            clips_to_concat.extend(video1, video2, transition)
            prev_video1 = video2
        clips_to_concat.append(prev_video1)

        # transitions = TransitionGenerator.get_transition_clips(videos, duration, transition_method)
        # videos_to_concat = []

        # for video, transition in zip(videos[:-1], transitions):
        #     videos_to_concat.append(video)
        #     videos_to_concat.append(transition)
        # videos_to_concat.append(videos[-1])

        return concatenate_videoclips(clips_to_concat)




from yta_multimedia.video.edition.effect.moviepy.position.objects.base_position_moviepy_effect import BasePositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_center import CoordinateCenter
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_corner import CoordinateCorner
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.position.utils.position import get_moviepy_position
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionSetPosition
from yta_general_utils.math.rate_functions import RateFunction
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, VideoClip
from typing import Union


# TODO: This should be removed when MoviepyVideoEffect
# is well prepared to handle 'apply' and
# 'apply_over_video'
class MoveWavePositionMoviepyEffect(BasePositionMoviepyEffect):
    """
    Effect of moving the element from one place in the screen (or outside
    of bounds) to another, moving like a wave, doing one arc above and
    one below.
    """
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], initial_position: Union[Position, CoordinateCorner, CoordinateCenter] = Position.OUT_LEFT, final_position: Union[Position, CoordinateCorner, CoordinateCenter] = Position.CENTER):
        """
        Applies the effect to the 'video' provided when initializing this
        effect class, and puts the video over a static black background
        image of 1920x1080.
        """
        return cls.apply_over_video(BasePositionMoviepyEffect.get_black_background_clip(video.duration), initial_position, final_position)

    def apply_over_video(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], background_video: Union[str, VideoClip, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], initial_position: Union[Position, CoordinateCorner, CoordinateCenter] = Position.OUT_LEFT, final_position: Union[Position, CoordinateCorner, CoordinateCenter] = Position.CENTER):
        """
        TODO: Write well

        Applies the effect on the video used when instantiating the
        effect, but applies the effect by placing it over the 
        'background_video' provided in this method (the 
        'background_video' will act as a background video for the 
        effect applied on the initial video).

        This method will set the video used when instantiating the
        effect as the most important, and its duration will be 
        considered as that. If the 'background_video' provided 
        has a duration lower than the original video, we will
        loop it to reach that duration. If the video is shorter
        than the 'background_video', we will crop the last one
        to fit the original video duration.
        """
        background_video = BasePositionMoviepyEffect.prepare_background_clip(background_video, video)

        BasePositionMoviepyEffect.validate_position(initial_position)
        BasePositionMoviepyEffect.validate_position(final_position)

        initial_position = get_moviepy_position(video, background_video, initial_position)
        final_position = get_moviepy_position(video, background_video, final_position)

        duration = video.duration
        # TODO: I could modify this to accept the 'rate_func' and apply it
        effect = video.set_position(lambda t: TFunctionSetPosition.wave(t, duration, initial_position, final_position, RateFunction.linear)).set_start(0).set_duration(duration)
        
        return CompositeVideoClip([
            background_video,
            effect
        ])


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, VideoClip
from typing import Union


class ScrollVideoEffect(VideoEffect):
    """
    This effect will make the clip be scrolled like if a zoomed
    region was surfing through the clip.
    """
    
    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], width = None, height = None, x_speed = None, y_speed = None, x_start = None, y_start = None):
        """
        Applies the effect to the provided 'video'.
        """
        video = VideoEffect.parse_moviepy_video(video)

        if width is None:
            width = 960

        if height is None:
            height = 540

        if x_speed is None:
            x_speed = 20

        if y_speed is None:
            y_speed = 20

        if x_start is None:
            x_start = 100

        if y_start is None:
            y_start = 100

        effect_name = VideoEffect.get_moviepy_vfx_effect('scroll')
        parameters = {}

        return video.fx(effect_name, **parameters).resize(video.size)


from yta_multimedia.video.frames import VideoFrameExtractor
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.image.edition.filter.sketch import image_to_sketch, image_to_line_sketch
from yta_general_utils.image.converter import ImageConverter
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ImageSequenceClip
from typing import Union
from types import FunctionType


def video_to_sketch_video(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], output_filename: Union[str, None]):
    # TODO: Document it
    return __video_to_frame_by_frame_filtered_video(video, image_to_sketch, output_filename)

def video_to_line_sketch_video(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], output_filename: Union[str, None]):
    """
    This method is very very slow. I should try to optimize it or just
    use not, because it doesn't make sense as a video.
    """
    # TODO: Document it
    return __video_to_frame_by_frame_filtered_video(video, image_to_line_sketch, output_filename)

def __video_to_frame_by_frame_filtered_video(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], filter_func: FunctionType, output_filename: Union[str, None] = None):
    """
    Internal function to be used by any of our video editing methods
    that actually use image filter frame by frame. They do the same
    by only changing the filter we apply.
    """
    # TODO: Check if 'filter_func' is a function
    video = VideoParser.to_moviepy(video)

    original_frames = VideoFrameExtractor.get_all_frames(video)
    sketched_frames = []
    for original_frame in original_frames:
        sketched_frames.append(ImageConverter.pil_image_to_numpy(filter_func(original_frame)))

    sketched_video = ImageSequenceClip(sketched_frames, fps = video.fps).set_audio(video.audio)

    if output_filename:
        sketched_video.write_videofile(output_filename)

    return sketched_video




from yta_multimedia.video.generation.manim.constants import HALF_SCENE_HEIGHT, HALF_SCENE_WIDTH
from yta_multimedia.video.generation.manim.utils.dimensions import ManimDimensions
from yta_general_utils.random import randrangefloat
from yta_general_utils.programming.enum import YTAEnum as Enum
from random import choice as randchoice


def get_center(video, background_video):
    """
    Returns the x,y coords in which the provided 'video' will
    be centered according to the provided 'background_video' in
    which it will be overlayed.

    This method returns two elements, first one is the x and the
    second one is the y.
    """
    # TODO: Ensure 'video' and 'background_video' are valid videos
    return background_video.w / 2 - video.w / 2, background_video.h / 2 - video.h / 2

class Position(Enum):
    """
    Enum class that represents a position inside the screen
    scene. This is used to position a video or an image inside
    the scene in an specific position defined by itself. It is
    useful with Manim and Moviepy video positioning and has
    been prepared to work with those engines.
    """
    OUT_TOP_LEFT = 'out_top_left'
    """
    Out of the screen, on the top left corner, just one pixel
    out of bounds.
    """
    IN_EDGE_TOP_LEFT = 'in_edge_top_left'
    """
    The center of the video is on the top left corner, so only
    the bottom right quarter part of the video is shown (inside
    the screen).
    """
    TOP_LEFT = 'top_left'
    """
    The video is completely visible, just at the top left 
    corner of the screen.
    """
    OUT_TOP = 'out_top'
    IN_EDGE_TOP = 'in_edge_top'
    TOP = 'top'
    OUT_TOP_RIGHT = 'out_top_right'
    IN_EDGE_TOP_RIGHT = 'in_edge_top_right'
    TOP_RIGHT = 'top_right'
    CENTER = 'center'
    OUT_RIGHT = 'out_right'
    IN_EDGE_RIGHT = 'in_edge_right'
    RIGHT = 'right'
    OUT_BOTTOM_RIGHT = 'out_bottom_right'
    IN_EDGE_BOTTOM_RIGHT = 'in_edge_bottom_right'
    BOTTOM_RIGHT = 'bottom_right'
    OUT_BOTTOM = 'out_bottom'
    IN_EDGE_BOTTOM = 'in_edge_bottom'
    BOTTOM = 'bottom'
    OUT_BOTTOM_LEFT = 'out_bottom_left'
    IN_EDGE_BOTTOM_LEFT = 'in_edge_bottom_left'
    BOTTOM_LEFT = 'bottom_left'
    OUT_LEFT = 'out_left'
    IN_EDGE_LEFT = 'in_edge_left'
    LEFT = 'left'

    HALF_TOP = 'half_top'
    HALF_TOP_RIGHT = 'half_top_right'
    HALF_RIGHT = 'half_right'
    HALF_BOTTOM_RIGHT = 'half_bottom_right'
    HALF_BOTTOM = 'half_bottom'
    HALF_BOTTOM_LEFT = 'half_bottom_left'
    HALF_LEFT = 'half_left'
    HALF_TOP_LEFT = 'half_top_left'

    QUADRANT_1_TOP_RIGHT_CORNER = 'quadrant_1_top_right_corner'
    QUADRANT_1_BOTTOM_RIGHT_CORNER = 'quadrant_1_bottom_right_corner'
    QUADRANT_1_BOTTOM_LEFT_CORNER = 'quadrant_1_bottom_left_corner'
    QUADRANT_2_TOP_LEFT_CORNER = 'quadrant_2_top_left_corner'
    QUADRANT_2_BOTTOM_RIGHT_CORNER = 'quadrant_2_bottom_right_corner'
    QUADRANT_2_BOTTOM_LEFT_CORNER = 'quadrant_2_bottom_left_corner'
    QUADRANT_3_TOP_RIGHT_CORNER = 'quadrant_3_top_right_corner'
    QUADRANT_3_TOP_LEFT_CORNER = 'quadrant_3_top_left_corner'
    QUADRANT_3_BOTTOM_LEFT_CORNER = 'quadrant_3_bottom_left_corner'
    QUADRANT_4_TOP_RIGHT_CORNER = 'quadrant_4_top_right_corner'
    QUADRANT_4_TOP_LEFT_CORNER = 'quadrant_4_top_left_corner'
    QUADRANT_4_BOTTOM_RIGHT_CORNER = 'quadrant_4_bottom_right_corner'

    RANDOM_INSIDE = 'random_inside'
    """
    A random position inside the screen with no pixels out of bounds.
    It is randomly chosen from one of all the options inside the limits
    we have.
    """
    RANDOM_OUTSIDE = 'random_outside'
    """
    A random position out of the screen limits. It is randomly chosen 
    from one of all the options outside the limits we have.
    """
    # TODO: Add more positions maybe (?)

    @classmethod
    def outside_positions_as_list(cls):
        """
        Returns the Position enums that are located out of
        the screen limits.
        """
        return [
            cls.OUT_TOP_LEFT,
            cls.OUT_TOP,
            cls.OUT_RIGHT,
            cls.OUT_BOTTOM_RIGHT,
            cls.OUT_BOTTOM,
            cls.OUT_BOTTOM_LEFT,
            cls.OUT_LEFT
        ]
    
    @classmethod
    def inside_positions_as_list(cls):
        """
        Returns the Position enums that are located inside
        the screen limits.
        """
        return list(set(cls.get_all()) - set(cls.outside_positions_as_list()) - set([cls.RANDOM_INSIDE]) - set([cls.RANDOM_OUTSIDE]))
    
    def get_manim_limits(self):
        """
        Return the left, right, top and bottom limits for this
        screen position. This edges represent the limits of the
        region in which the video should be placed to fit this
        screen position.

        We consider each screen region as a limited region of
        half of the scene width and height.

        Corner limits:
        [-7-1/9,  4, 0]   [0,  4, 0]   [7+1/9,  4, 0]
        [-7-1/9,  0, 0]   [0,  0, 0]   [7+1/9,  0, 0]
        [-7-1/9, -4, 0]   [0, -4, 0]   [7+1/9, -4, 0]
        """
        # TODO: I think I should consider regions of 1/8 of width and height
        # so 1 quadrant is divided into 4 pieces and I build all the positions
        # for also those quadrants
        # TODO: I'm missing the QUADRANT_1_HALF_TOP and HALF_TOP_OUT, ...
        if self == Position.TOP:
            return -HALF_SCENE_WIDTH / 2, HALF_SCENE_WIDTH / 2, HALF_SCENE_HEIGHT, 0
        elif self == Position.BOTTOM:
            return -HALF_SCENE_WIDTH / 2, HALF_SCENE_WIDTH / 2, -HALF_SCENE_HEIGHT, 0
        elif self == Position.LEFT:
            return -HALF_SCENE_WIDTH, 0, HALF_SCENE_HEIGHT / 2, -HALF_SCENE_HEIGHT / 2
        elif self == Position.RIGHT:
            return 0, HALF_SCENE_WIDTH, HALF_SCENE_HEIGHT / 2, -HALF_SCENE_HEIGHT / 2
        # TODO: Add missing

        # TODO: Is this method necessary (?)

    def get_manim_position(self, width: float, height: float):
        """
        Return the position in which the mobject must be placed to
        be exactly in this position.

        This method returns a x,y,z position.
        """
        # TODO: 'width' and 'height' must be manim

        if self == Position.TOP:
            return 0, HALF_SCENE_HEIGHT - height / 2, 0
        elif self == Position.TOP_RIGHT:
            return HALF_SCENE_WIDTH - width / 2, HALF_SCENE_HEIGHT - height / 2, 0
        elif self == Position.RIGHT:
            return HALF_SCENE_WIDTH - width / 2, 0, 0
        elif self == Position.BOTTOM_RIGHT:
            return HALF_SCENE_WIDTH - width / 2, -HALF_SCENE_HEIGHT + height / 2, 0
        elif self == Position.BOTTOM:
            return 0, -HALF_SCENE_HEIGHT + height / 2, 0
        elif self == Position.BOTTOM_LEFT:
            return -HALF_SCENE_WIDTH + width / 2, -HALF_SCENE_HEIGHT + height / 2, 0
        elif self == Position.LEFT:
            return -HALF_SCENE_WIDTH + width / 2, 0, 0
        elif self == Position.TOP_LEFT:
            return -HALF_SCENE_WIDTH + width / 2, HALF_SCENE_HEIGHT - height / 2, 0
        elif self == Position.CENTER:
            return 0, 0, 0
        elif self == Position.IN_EDGE_TOP_LEFT:
            return -HALF_SCENE_WIDTH, HALF_SCENE_HEIGHT, 0
        elif self == Position.OUT_TOP_LEFT:
            return -HALF_SCENE_WIDTH - width / 2, HALF_SCENE_HEIGHT + height / 2, 0
        elif self == Position.IN_EDGE_TOP:
            return 0, HALF_SCENE_HEIGHT, 0
        elif self == Position.OUT_TOP:
            return 0, HALF_SCENE_HEIGHT + height / 2, 0
        elif self == Position.IN_EDGE_TOP_RIGHT:
            return HALF_SCENE_WIDTH, HALF_SCENE_HEIGHT, 0
        elif self == Position.OUT_TOP_RIGHT:
            return HALF_SCENE_WIDTH + width / 2, HALF_SCENE_HEIGHT + height / 2, 0
        elif self == Position.IN_EDGE_RIGHT:
            return HALF_SCENE_WIDTH, 0, 0
        elif self == Position.OUT_RIGHT:
            return HALF_SCENE_WIDTH + width / 2, 0, 0
        elif self == Position.IN_EDGE_BOTTOM_RIGHT:
            return HALF_SCENE_WIDTH, -HALF_SCENE_HEIGHT, 0
        elif self == Position.OUT_BOTTOM_RIGHT:
            return HALF_SCENE_WIDTH + width / 2, -HALF_SCENE_HEIGHT - height / 2, 0
        elif self == Position.IN_EDGE_BOTTOM:
            return 0, -HALF_SCENE_HEIGHT, 0
        elif self == Position.OUT_BOTTOM:
            return 0, -HALF_SCENE_HEIGHT - height / 2, 0
        elif self == Position.IN_EDGE_BOTTOM_LEFT:
            return -HALF_SCENE_WIDTH, -HALF_SCENE_HEIGHT, 0
        elif self == Position.OUT_BOTTOM_LEFT:
            return -HALF_SCENE_WIDTH - width / 2, -HALF_SCENE_HEIGHT - height / 2, 0
        elif self == Position.IN_EDGE_LEFT:
            return -HALF_SCENE_WIDTH, 0, 0
        elif self == Position.OUT_LEFT:
            return -HALF_SCENE_WIDTH - width / 2, 0, 0
        elif self == Position.HALF_TOP:
            return 0, HALF_SCENE_HEIGHT / 2, 0
        elif self == Position.HALF_TOP_RIGHT:
            return HALF_SCENE_WIDTH / 2, HALF_SCENE_HEIGHT / 2, 0
        elif self == Position.HALF_RIGHT:
            return HALF_SCENE_WIDTH / 2, 0, 0
        elif self == Position.HALF_BOTTOM_RIGHT:
            return HALF_SCENE_WIDTH / 2, -HALF_SCENE_HEIGHT / 2, 0
        elif self == Position.HALF_BOTTOM:
            return 0, -HALF_SCENE_HEIGHT / 2, 0
        elif self == Position.HALF_BOTTOM_LEFT:
            return -HALF_SCENE_WIDTH / 2, -HALF_SCENE_HEIGHT / 2, 0
        elif self == Position.HALF_LEFT:
            return -HALF_SCENE_WIDTH / 2, 0, 0
        elif self == Position.HALF_TOP_LEFT:
            return -HALF_SCENE_WIDTH / 2, HALF_SCENE_HEIGHT / 2, 0
        elif self == Position.QUADRANT_1_TOP_RIGHT_CORNER:
            return -width / 2, HALF_SCENE_HEIGHT - height / 2, 0
        elif self == Position.QUADRANT_1_BOTTOM_RIGHT_CORNER:
            return -width / 2, height / 2, 0
        elif self == Position.QUADRANT_1_BOTTOM_LEFT_CORNER:
            return -HALF_SCENE_WIDTH + width / 2, height / 2, 0
        elif self == Position.QUADRANT_2_TOP_LEFT_CORNER:
            return width / 2, HALF_SCENE_HEIGHT - height / 2, 0
        elif self == Position.QUADRANT_2_BOTTOM_RIGHT_CORNER:
            return HALF_SCENE_WIDTH - width / 2, height / 2, 0
        elif self == Position.QUADRANT_2_BOTTOM_LEFT_CORNER:
            return width / 2, height / 2, 0
        elif self == Position.QUADRANT_3_TOP_LEFT_CORNER:
            return width / 2, -height / 2, 0
        elif self == Position.QUADRANT_3_TOP_RIGHT_CORNER:
            return HALF_SCENE_WIDTH - width / 2, -height / 2, 0
        elif self == Position.QUADRANT_3_BOTTOM_LEFT_CORNER:
            return width / 2, -HALF_SCENE_HEIGHT + height / 2, 0
        elif self == Position.QUADRANT_4_TOP_LEFT_CORNER:
            return -HALF_SCENE_WIDTH + width / 2, -height / 2, 0
        elif self == Position.QUADRANT_4_TOP_RIGHT_CORNER:
            return -width / 2, -height / 2, 0
        elif self == Position.QUADRANT_4_BOTTOM_RIGHT_CORNER:
            return -width / 2, -HALF_SCENE_HEIGHT + height / 2, 0
        elif self == Position.RANDOM_INSIDE:
            return randchoice(Position.inside_positions_as_list()).get_manim_position(width, height)
        elif self == Position.RANDOM_OUTSIDE:
            return randchoice(Position.outside_positions_as_list()).get_manim_position(width, height)
        
    def get_manim_random_position(self, width: float, height: float):
        """
        Calculate the position in which the center of the element 
        with the provided 'width' and 'height' must be placed to
        obtain this position.

        The position will be provided as a 3D vector but the z
        axis will be always 0.

        The provided 'width' and 'height' must be in pixels.
        """
        # TODO: Check if the provided 'width' and 'height' are in
        # in pixels and valid or if not

        # TODO: By now I'm just using the limits as numeric limits
        # for random position that will be used as the center of the
        # video, but we will need to consider the video dimensions 
        # in a near future to actually position it well, because the
        # video can be out of the scene right now with this approach
        left, right, top, bottom = self.get_manim_limits()

        x, y = randrangefloat(left, right, ManimDimensions.width_to_manim_width(1)), randrangefloat(top, bottom, ManimDimensions.height_to_manim_height(1))

        # If video is larger than HALF/2 it won't fit correctly.
        if width > HALF_SCENE_WIDTH or height > HALF_SCENE_HEIGHT:
            # TODO: Video is bigger than the region, we cannot make
            # it fit so... what can we do (?)
            return x, y

        if x - width / 2 < left:
            x += left - (x - width / 2)
        if x + width / 2 > right:
            x -= (x + width / 2) - right
        if y - height / 2 < bottom:
            y += bottom - (y - height / 2)
        if y + height / 2 > top:
            y -= (y + height / 2) - top

        return x, y, 0
        # TODO: Is this method necessary (?) I think yes, because
        # it generates a random position in a region, not a random
        # choice of our predefined (and always the same) positions

    def get_moviepy_position(self, video, background_video):
        """
        This method will calculate the (x, y) tuple position for the provided
        'video' over the also provided 'background_video' that would be,
        hypothetically, a 1920x1080 black color background static image. The
        provided 'position' will be transformed into the (x, y) tuple according
        to our own definitions.
        """
        # TODO: Do 'video' and 'background_video' checkings
        position_tuple = None

        if self == Position.CENTER:
            position_tuple = (get_center(video, background_video))

        #           Edges below
        # TOP
        elif self == Position.OUT_TOP:
            position_tuple = ((background_video.w / 2) - (video.w / 2), -video.h)
        elif self == Position.IN_EDGE_TOP:
            position_tuple = ((background_video.w / 2) - (video.w / 2), -(video.h / 2))
        elif self == Position.TOP:
            position_tuple = ((background_video.w / 2) - (video.w / 2), 0)
        # TOP RIGHT
        elif self == Position.OUT_TOP_RIGHT:
            position_tuple = (background_video.w, -video.h)
        elif self == Position.IN_EDGE_TOP_RIGHT:
            position_tuple = (background_video.w - (video.w / 2), -(video.h / 2))
        elif self == Position.TOP_RIGHT:
            position_tuple = (background_video.w - video.w, 0)
        # RIGHT
        elif self == Position.OUT_RIGHT:
            position_tuple = (background_video.w, (background_video.h / 2) - (video.h / 2))
        elif self == Position.IN_EDGE_RIGHT:
            position_tuple = (background_video.w - (video.w / 2), (background_video.h / 2) - (video.h / 2))
        elif self == Position.RIGHT:
            position_tuple = (background_video.w - video.w, (background_video.h / 2) - (video.h / 2))
        # BOTTOM RIGHT
        elif self == Position.OUT_BOTTOM_RIGHT:
            position_tuple = (background_video.w, background_video.h)
        elif self == Position.IN_EDGE_BOTTOM_RIGHT:
            position_tuple = (background_video.w - (video.w / 2), background_video.h - (video.h / 2))
        elif self == Position.BOTTOM_RIGHT:
            position_tuple = (background_video.w - video.w, background_video.h - video.h)
        # BOTTOM
        elif self == Position.OUT_BOTTOM:
            position_tuple = ((background_video.w / 2) - (video.w / 2), background_video.h)
        elif self == Position.IN_EDGE_BOTTOM:
            position_tuple = ((background_video.w / 2) - (video.w / 2), background_video.h - (video.h / 2))
        elif self == Position.BOTTOM:
            position_tuple = ((background_video.w / 2) - (video.w / 2), background_video.h - video.h)
        # BOTTOM LEFT
        elif self == Position.OUT_BOTTOM_LEFT:
            position_tuple = (-video.w, background_video.h)
        elif self == Position.IN_EDGE_BOTTOM_LEFT:
            position_tuple = (-(video.w / 2), background_video.h - (video.h / 2))
        elif self == Position.BOTTOM_LEFT:
            position_tuple = (0, background_video.h - video.h)
        # LEFT
        elif self == Position.OUT_LEFT:
            position_tuple = (-video.w, (background_video.h / 2) - (video.h / 2))
        elif self == Position.IN_EDGE_LEFT:
            position_tuple = (-(video.w / 2), (background_video.h / 2) - (video.h / 2))
        elif self == Position.LEFT:
            position_tuple = (0, (background_video.h / 2) - (video.h / 2))
        # TOP LEFT
        elif self == Position.OUT_TOP_LEFT:
            position_tuple = (-video.w, -video.h)
        elif self == Position.IN_EDGE_TOP_LEFT:
            position_tuple = (-(video.w / 2), -(video.h / 2))
        elif self == Position.TOP_LEFT:
            position_tuple = (0, 0)

        # HALF POSITIONS
        elif self == Position.HALF_TOP:
            position_tuple = (background_video.w / 2 - video.w / 2, background_video.h / 4 - video.h / 2)
        elif self == Position.HALF_RIGHT:
            position_tuple = (3 * background_video.w / 4 - video.w / 2, background_video.h / 2 - video.h / 2)
        elif self == Position.HALF_BOTTOM:
            position_tuple = (background_video.w / 2 - video.w / 2, 3 * background_video.h / 4 - video.h / 2)
        elif self == Position.HALF_LEFT:
            position_tuple = (background_video.w / 4 - video.w / 2, background_video.h / 2 - video.h / 2)
        elif self == Position.HALF_TOP_RIGHT:
            position_tuple = (3 * background_video.w / 4 - video.w / 2, background_video.h / 4 - video.h / 2)
        elif self == Position.HALF_BOTTOM_RIGHT:
            position_tuple = (3 * background_video.w / 4 - video.w / 2, 3 * background_video.h / 4 - video.h / 2)
        elif self == Position.HALF_BOTTOM_LEFT:
            position_tuple = (background_video.w / 4 - video.w / 2, 3 * background_video.h / 4 - video.h / 2)
        elif self == Position.HALF_TOP_LEFT:
            position_tuple = (background_video.w / 4 - video.w / 2, background_video.h / 4 - video.h / 2)

        # QUADRANT CORNERS
        elif self == Position.QUADRANT_1_TOP_RIGHT_CORNER:
            position_tuple = (background_video.w / 2 - video.w, 0)
        elif self == Position.QUADRANT_1_BOTTOM_LEFT_CORNER:
            position_tuple = (0, background_video.h / 2 - video.h)
        elif self == Position.QUADRANT_1_BOTTOM_RIGHT_CORNER:
            position_tuple = (background_video.w / 2 - video.w, background_video.h / 2 - video.h)
        elif self == Position.QUADRANT_2_BOTTOM_LEFT_CORNER:
            position_tuple = (background_video.w / 2, background_video.h / 2 - video.h)
        elif self == Position.QUADRANT_2_BOTTOM_RIGHT_CORNER:
            position_tuple = (background_video.w - video.w, background_video.h / 2 - video.h)
        elif self == Position.QUADRANT_2_TOP_LEFT_CORNER:
            position_tuple = (background_video.w / 2, 0)
        elif self == Position.QUADRANT_3_BOTTOM_LEFT_CORNER:
            position_tuple = (background_video.w / 2, background_video.h - video.h)
        elif self == Position.QUADRANT_3_TOP_LEFT_CORNER:
            position_tuple = (background_video.w / 2, background_video.h / 2)
        elif self == Position.QUADRANT_3_TOP_RIGHT_CORNER:
            position_tuple = (background_video.w - video.w, background_video.h / 2)
        elif self == Position.QUADRANT_4_BOTTOM_RIGHT_CORNER:
            position_tuple = (background_video.w / 2 - video.w, background_video.h - video.h)
        elif self == Position.QUADRANT_4_TOP_LEFT_CORNER:
            position_tuple = (0, background_video.h / 2)
        elif self == Position.QUADRANT_4_TOP_RIGHT_CORNER:
            position_tuple = (background_video.w / 2 - video.w, background_video.h / 2)

        # RANDOMs
        elif self == Position.RANDOM_INSIDE:
            return randchoice(Position.inside_positions_as_list()).get_moviepy_position(video, background_video)
        elif self == Position.RANDOM_OUTSIDE:
            return randchoice(Position.outside_positions_as_list()).get_moviepy_position(video, background_video)

        return position_tuple


from yta_multimedia.video.edition.effect.moviepy.objects import MoviepySetPosition, MoviepyRotate, MoviepyResize
from yta_multimedia.video.edition.effect.moviepy.position.objects.base_position_moviepy_effect import BasePositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.utils.position import get_moviepy_position
from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from moviepy.editor import CompositeVideoClip, VideoFileClip, ImageClip, VideoClip, ColorClip
from typing import Union


class MoviepyVideoEffect(VideoEffect):
    """
    Class to apply effect on moviepy video by providing
    MoviepyZoom, MoviepyRotation and/or MoviepyPosition
    arguments.
    """
    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], resize: MoviepyResize = None, set_position: MoviepySetPosition = None, rotate: MoviepyRotate = None):
        """
        Apply the effect on the provided 'video'.

        :param MoviepyResize resize: The resize effect to apply (as a lambda t function).

        :param MoviepySetPosition set_position: The position to apply (as a lambda t function).

        :param MoviepyRotate rotate: The rotation effect to apply (as a lambda t function).
        """
        return cls.apply_over_video(video = video, background_video = BasePositionMoviepyEffect.get_black_background_clip(video.duration), resize = resize, set_position = set_position, rotate = rotate)
        # TODO: Remove this below when checked
        # video = VideoEffect.parse_moviepy_video(video)
        # # We reset the mask to avoid problems with zoom
        # video = video.add_mask()

        # # TODO: Check 'resize' is instance of MoviepyResize or valid value
        # # TODO: Check 'set_position' is instance of MoviepySetPosition or valid value
        # # TODO: Check 'rotate' is instance of MoviepyRotate or valid value

        # fps = video.fps
        # duration = video.duration
        # screensize = video.size

        # # Basic configuration (is position needed here (?))
        # effected_video = video.resize(screensize).set_position(('center', 'center')).set_duration(duration).set_fps(fps)

        # if resize:
        #     effected_video = effected_video.resize(lambda t: MoviepyResize.t_function(t, duration, resize.initial_size, resize.final_size, resize.rate_func))
        # if set_position:
        #     effected_video = effected_video.set_position(lambda t: MoviepySetPosition.t_function(t, duration, set_position.initial_position, set_position.final_position, set_position.rate_func))
        # if rotate:
        #     effected_video = effected_video.rotate(lambda t: MoviepyRotate.t_function(t, duration, rotate.initial_rotation, rotate.final_rotation, rotate.rate_func))

        # return CompositeVideoClip([effected_video], size = screensize)
    
    @classmethod
    def apply_over_video(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], background_video: Union[str, VideoClip, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], resize: MoviepyResize = None, set_position: MoviepySetPosition = None, rotate: MoviepyRotate = None):
        video = VideoEffect.parse_moviepy_video(video, True)
        background_video = VideoEffect.parse_moviepy_video(background_video)
        background_video = BasePositionMoviepyEffect.prepare_background_clip(background_video, video)

        fps = video.fps
        duration = video.duration
        screensize = video.size

        effected_video = video.resize(screensize).set_position(('center', 'center')).set_duration(duration).set_fps(fps)

        # # TODO: Check 'resize' is instance of MoviepyResize or valid value
        # # TODO: Check 'set_position' is instance of MoviepySetPosition or valid value
        # # TODO: Check 'rotate' is instance of MoviepyRotate or valid value

        if resize:
            effected_video = effected_video.resize(lambda t: resize.t_function(t, duration, resize.initial_size, resize.final_size, resize.rate_func))

        if set_position:
            BasePositionMoviepyEffect.validate_position(set_position.initial_position)
            BasePositionMoviepyEffect.validate_position(set_position.final_position)

            effected_video = effected_video.set_position(lambda t: set_position.t_function(t, duration, get_moviepy_position(video, background_video, set_position.initial_position), get_moviepy_position(video, background_video, set_position.final_position), set_position.rate_func))

        if rotate:
            effected_video = effected_video.rotate(lambda t: rotate.t_function(t, duration, rotate.initial_rotation, rotate.final_rotation, rotate.rate_func))

        return CompositeVideoClip([
            background_video,
            effected_video
        ])


from yta_multimedia.video.parser import VideoParser
from yta_general_utils.dimensions import adjust_to_aspect_ratio
from moviepy.editor import VideoClip
from moviepy.video.fx.all import crop
from typing import Union


def resize_video(video: Union[str, VideoClip], size, output_filename: Union[str, None] = None):
    """
    Resizes the video to the provided 'size' by cropping a
    region of the given 'video' that fits the 'size' aspect
    ratio and resizing that region to the 'size'.

    This method is using the whole video and then resizing,
    so the quality of the video is preserved and no small
    regions are used. The most part of the video is 
    preserved.

    This method returns the video modified.

    This method will write the video if 'output_filename' is
    provided.
    """
    video = VideoParser.to_moviepy(video)

    # We need to adjust the video to the AR of the given 'size'
    new_ar = size[0] / size[1]
    video_width, video_height = video.size
    new_video_size = adjust_to_aspect_ratio(video_width, video_height, new_ar)

    video = crop(video, width = new_video_size[0], height = new_video_size[1], x_center = video_width / 2, y_center = video_height / 2)

    # We resize it (up or down) to the desired 'size'
    video = video.resize((size[0], size[1]))

    if output_filename:
        video.write_videofile(output_filename)

    return video


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, VideoClip
from typing import Union


class FadeInVideoEffect(VideoEffect):
    """
    This effect will make the video appear progressively
    lasting the provided 'duration' time or the whole 
    clip time duration if None 'duration' provided.

    The 'color' provided must be an array containing
    the rgb colors (default is [0, 0, 0], which is black).
    """

    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], duration = None, color = None):
        """
        Applies the effect to the provided 'video'.
        """
        video = VideoEffect.parse_moviepy_video(video)

        if duration is None:
            duration = video.duration

        # TODO: Apply new Color class when available to parse color parameters
        if color is None:
            color = [0, 0, 0]

        effect_name = VideoEffect.get_moviepy_vfx_effect('fadein')
        parameters = {
            'duration': duration,
            'initial_color': color
        }

        return video.fx(effect_name, **parameters)




from yta_multimedia.image.edition.filter.gameboy import image_file_to_gameboy
from yta_multimedia.image.edition.filter.pixelate import pixelate_image
from yta_multimedia.image.edition.filter.sticker import image_file_to_sticker


class ImageFilter:
    """
    Class to simplify and encapsulate the functionality related to
    applying filters to an image.
    """
    @staticmethod
    def to_gameboy(image_filename: str, output_filename: str):
        """
        Apply the original GameBoy colors palette to the provided
        'image_filename'.
        """
        # TODO: We need to refactor this when the original method
        # is also refactored
        return image_file_to_gameboy(image_filename, output_filename)
    
    @staticmethod
    def pixelate(image, i_size, output_filename: str):
        """
        Apply a pixelation of 'i_size' to the provided 'image'.
        """
        # TODO: Check that it is a valid image
        # TODO: Validate the provided 'i_size'
        return pixelate_image(image, i_size, output_filename)
    
    @staticmethod
    def to_sticker(image_filename: str, output_filename: str):
        """
        Turn the provided 'image_filename' into a sticker, which
        is the same image without the background and with a new
        white and wide border.
        """
        # TODO: Check that it is a valid 'image_filename'
        # TODO: Refactor to accept other image types, not only files
        return image_file_to_sticker(image_filename, output_filename)
    
    # TODO: Add 'image_to_sketch' if working, but I think it is
    # too heavy processing for nothing interesting...


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from yta_multimedia.video.frames import VideoFrameExtractor
from moviepy.editor import CompositeVideoClip, VideoFileClip, ImageClip, concatenate_videoclips, VideoClip, ColorClip
from typing import Union


class StopMotionVideoEffect(VideoEffect):
    """
    Creates a Stop Motion effect in the provided video by dropping the frames
    per second but maintaining the original frames ratio.
    """
    
    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip]):
        """
        Applies the effect on the provided 'video'.
        """
        video = VideoEffect.parse_moviepy_video(video)

        FRAMES_TO_JUMP = 5

        clips = []
        for frame_number in range((int) (video.fps * video.duration)):
            if frame_number % FRAMES_TO_JUMP == 0:
                frame = VideoFrameExtractor.get_frame_by_frame_number(video, frame_number)
                clips.append(ImageClip(frame, duration = FRAMES_TO_JUMP / video.fps).set_fps(video.fps))

        return concatenate_videoclips(clips).set_audio(video.audio).set_fps(video.fps)


    


"""
This is an adaption of the rate functions found in manim
library for the moviepy library. I have to use lambda t
functions with the current frame time to be able to resize
or reposition a video, so I have adapted the manim rate
functions to be able to return a factor that, with specific
moviepy functions, will make a change with the factor that
has been calculated with the corresponding rate function.

You can see 'manim/utils/rate_functions.py'.

This is the way I have found to make it work and to be able
to build smoother animations. As manim docummentation says,
the rate functions have been inspired by the ones listed in
this web page: https://easings.net/
"""
from yta_general_utils.math.rate_functions import RateFunction
import numpy as np

from math import pi


# TODO: This method could be private
def rate_function(t: float, duration: float, rate_func: type, *args, **kwargs):
    """
    You need to provide one of the functions of RateFunction class
    as the 'rate_func' parameter to be able to make it work, and 
    pass the needed args to it.
    """
    return rate_func(t / duration, *args, **kwargs)

class TFunctionResize:
    """
    Class to simplify and encapsulate the functionality
    related with the zoom of a video (by resizing it).

    You can use these methods to make a moviepy video
    rotate from the 'initial_size' to the 'final_size'
    using the 'rate_func'. The original video size is 1.
    """
    @staticmethod
    def resize_from_to(t: float, duration: float, initial_size: float, final_size: float, rate_func: type = RateFunction.linear, *args, **kwargs):
        """
        A function to be applied in moviepy '.resize()' as 'lambda t:
        zoom_from_to()'.

        A size value of 1 means no resize, while a value greater than 1
        is a zoom_in effect (enlarging it), and a value lower than 1 is
        a zoom_out effect (shortening it). The size must be always
        greater than 0.

        The 'rate_func' must be one of the existing functions in the
        yta_general_utils.math.rate_functions.RateFunction class.
        """
        # TODO: Check 'initial_size', 'final_size' and 'rate_func'
        # are valid
        return initial_size + (final_size - initial_size) * rate_function(t, duration, rate_func, *args, **kwargs)

class TFunctionRotate:
    """
    Class to simplify and encapsulate the functionality
    related with the rotation of a video.

    You can use these methods to make a moviepy video
    rotate from the 'initial_rotation' to the 
    'final_rotation' using the 'rate_func'.
    """
    @staticmethod
    def rotate_from_to(t: float, duration: float, initial_rotation: int, final_rotation: int, rate_func: type = RateFunction.linear, *args, **kwargs):
        """
        A function to be applied in moviepy '.rotate()' as 'lambda
        t: rotate_from_to()'.

        The 'initial_rotation' and the 'final_rotation' must be, each
        one, the expected rotation, expressed in degrees.

        The 'rate_func' must be one of the existing functions in the
        yta_general_utils.math.rate_functions.RateFunction class.
        """
        # TODO: Check 'initial_rotation', 'final_rotation' and 
        # 'rate_func' are valid
        return initial_rotation + (final_rotation - initial_rotation) * rate_function(t, duration, rate_func, *args, *kwargs)

# TODO: I don't know where I should place this, but
# I keep it here by now as it is related with the
# t functions
# TODO: This below is related to positioning. As we
# know, we can set a rate_func for the movement, but
# the path has been always a straight line. Now this
# changes with the functions below:
class TFunctionSetPosition:
    """
    Class to simplify and encapsulate the functionality
    related with the path to follow between the point
    from which we want to go and where we want to reach.

    You can use these methods to make a moviepy video go
    from the 'initial_position' to the 'final_position'
    using the 'rate_func' but also choosing the path that
    the method is defining.
    """
    def circles_at_position(t: float, duration: float, position: tuple, radius: int = 200, time_per_circle: float = 1, rate_func: type = RateFunction.linear):
        """
        Returns the (x, y) position tuple for the moviepy '.set_position()' effect,
        for each 't' provided, that will make the element move in circles with the
        provided 'radius'. The 'radius' parameter is the distance between the origin
        and the path the clip will follow. The 'cicle_time' is the time (in seconds)
        needed for a complete circle to be completed by the movement.

        If you provide the video duration as 'cicle_time', the video will make only
        one whole circle
        """
        # TODO: Check this
        circle_factor = rate_function(t % time_per_circle, time_per_circle, rate_func)

        # TODO: Do checkings
        return position[0] + radius * np.cos((circle_factor) * 2 * pi), position[1] + radius * np.sin((circle_factor) * 2 * pi)
        # Working code was this below:
        # return x + radius * np.cos((t / time_per_circle) * 2 * math.pi), y + radius * np.sin((t / time_per_circle) * 2 * math.pi)

    def linear(t: float, duration: float, initial_position: tuple, final_position: tuple, rate_func: type = RateFunction.linear):
        factor = rate_function(t, duration, rate_func)

        x = initial_position[0] + (final_position[0] - initial_position[0]) * factor
        y = initial_position[1] + (final_position[1] - initial_position[1]) * factor

        return (x, y)
    
    def sinusoidal(t: float, duration: float, initial_position: tuple, final_position: tuple, rate_func: type = RateFunction.linear):
        """
        Returns the (x, y) position tuple for the moviepy '.set_position()' method,
        for each 't' provided, that will make the element move doing a sinusoidal
        movement from the 'initial_position' to the 'final_position' in the provided
        'duration' time.
        """
        sinusoidal_wave_amplitude = 100
        sinusoidal_wave_frequency = 2
        factor = rate_function(t, duration, rate_func)

        x = initial_position[0] + factor * (final_position[0] - initial_position[0])
        y = initial_position[1] + factor * (final_position[1] - initial_position[1]) + sinusoidal_wave_amplitude * np.sin(2 * np.pi * sinusoidal_wave_frequency * factor)

        return (x, y)
    
    def zigzag(t: float, duration: float, initial_position: tuple, final_position: tuple, rate_func: type = RateFunction.linear):
        factor = rate_function(t, duration, rate_func)
        amplitude = 15
        
        x = initial_position[0] + factor * (final_position[0] - initial_position[0])
        y = initial_position[1] + factor * (final_position[1] - initial_position[1]) + amplitude * np.sin(10 * np.pi * factor)

        return (x, y)
    
    def wave(t: float, duration: float, initial_position: tuple, final_position: tuple, rate_func: type = RateFunction.linear):
        """
        Returns the (x, y) position tuple for the moviepy '.set_position()'
        method, for each 't' provided, that will make the element go to the
        'initial_position' to the 'final_position' by doing two arcs (a 
        wave), one top and one bottom.
        """
        # TODO: Fix the 'y', it is not starting from the initial 
        # point
        sinusoidal_wave_amplitude = 200  # Maximum height of the arc
        sinusoidal_wave_frequency = 2 * np.pi / (final_position[0] - initial_position[0])  # Controls the frequency of the sine wave
        offset = (final_position[0] - initial_position[0]) / 2  # Center of the arc
        factor = rate_function(t, duration, rate_func)
        x = initial_position[0] + (final_position[0] - initial_position[0]) * factor

        # Sinusoidal curve: y = amplitude * sin(frequency * (x - offset)) + average_y
        average_y = (initial_position[1] + final_position[1]) / 2
        y = sinusoidal_wave_amplitude * np.sin(sinusoidal_wave_frequency * (x - initial_position[0] - offset)) + average_y
        
        return (x, y)
    
    def arc(t: float, duration: float, initial_position: tuple, final_position: tuple, arc_is_bottom: bool = False, rate_func: type = RateFunction.linear):
        """
        Returns the (x, y) position tuple for the moviepy '.set_position()'
        method, for each 't' provided, that will make the element go to the
        'initial_position' to the 'final_position' by doing one single 
        arc movement. The movement will be an arc above, or bottom if the
        'arc_is_bottom' parameter is set as True. The 'max_height' parameter
        is to set the maximum height of the arc.
        """
        # TODO: Fix the 'y', it is not starting from the initial 
        # point
        max_height = 300
        factor = rate_function(t, duration, rate_func)
        x = initial_position[0] + (final_position[0] - initial_position[0]) * factor

        # Calculate the midpoint and the maximum y value
        midpoint_x = (initial_position[0] + final_position[0]) / 2
        midpoint_y = (initial_position[1] + final_position[1]) / 2
        if arc_is_bottom:
            midpoint_y += max_height
        else:
            midpoint_y -= max_height

        # Calculate the arc movement:   y = a(x - h)^2 + k
        if arc_is_bottom:
            a = 4 * max_height / ((final_position[0] - initial_position[0])**2)
            y = -a * (x - midpoint_x)**2 + midpoint_y
        else:
            a = -4 * max_height / ((final_position[0] - initial_position[0])**2)
            y = -a * (x - midpoint_x)**2 + midpoint_y
        
        return (x, y)
    
    def circles(t: float, duration: float, initial_position: tuple, final_position: tuple, rate_func: type = RateFunction.linear):
        # TODO: Fix the 'y', it is not starting from the initial 
        # point
        from math import sqrt, atan2, cos, sin

        factor = rate_function(t, duration, rate_func)
        # Desempaquetar las coordenadas iniciales y finales
        x1, y1 = initial_position
        x2, y2 = final_position

        # Calcular el vector de dirección
        dx = x2 - x1
        dy = y2 - y1
        distancia = sqrt(dx**2 + dy**2)

        # Calcular el centro del círculo
        # El centro está a medio camino de A y B, pero desplazado perpendicularmente
        # La dirección perpendicular al vector AB
        perpendicular_dx = -dy
        perpendicular_dy = dx
        
        # Normalizar el vector perpendicular
        longitud_perpendicular = sqrt(perpendicular_dx**2 + perpendicular_dy**2)
        perpendicular_dx /= longitud_perpendicular
        perpendicular_dy /= longitud_perpendicular
        
        # El radio del círculo es la mitad de la distancia entre A y B
        radio = distancia / 2
        
        # El centro del círculo está a la mitad de la distancia entre A y B y desplazado en la dirección perpendicular
        cx = (x1 + x2) / 2 + radio * perpendicular_dx
        cy = (y1 + y2) / 2 + radio * perpendicular_dy

        # El ángulo de inicio del movimiento (desde A)
        angulo_inicio = atan2(y1 - cy, x1 - cx)
        
        # El ángulo final del movimiento (hacia B)
        angulo_final = atan2(y2 - cy, x2 - cx)
        
        # Hacer que el objeto se mueva de A a B a través del círculo en función de t (que va de 0 a 1)
        # Convertir t en un ángulo dentro del círculo
        angulo_actual = angulo_inicio + (angulo_final - angulo_inicio) * factor
        
        # Coordenadas del objeto en la trayectoria circular
        x = cx + radio * cos(angulo_actual)
        y = cy + radio * sin(angulo_actual)
        
        return (x, y)

    


"""
Using the 'get_frame' in moviepy returns a numpy array of 3 or 1 
dimension depending on the type of clip. If the clip is a main 
clip, a 3d array with values between [0, 255] is returned (one
[255, 255, 255] array would be a white pixel). If the clip is a
mask clip, a 1d array with values between [0, 1] is returned (the
1 completely transparent, the 0 is not).

Thats why you need to normalize or denormalize those values to
work with them because they are different and turning frames into
an image would need you to have the same range [0, 1] or [0, 255].

Check the Pillow, ImageIO and other libraries to see what kind of
numpy arrays are needed to write (or read) images.
"""
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.image.converter import ImageConverter
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import NumberValidator
from imageio import imsave

import numpy as np


class FrameExtractionType(Enum):
    TIME = 'time'
    FRAME_NUMBER = 'frame_number'

class VideoFrameExtractor:
    """
    Class to simplify the process of extracting video frames by time
    or frame number.
    """
    # TODO: Implement the 'output_filename' optional parameter to store
    # the frames if parameter is provided.
    @classmethod
    def _get_frame_time(cls, video, frame_number_or_time, extraction_type: FrameExtractionType):
        """
        Returns the time to obtain the frame of the provided 'video'
        according to the provided 'frame_number_or_time' parameter and
        the desired 'extraction_type'.

        This method is useful to be used first to validate that the
        frame times or numbers are valid and avoid getting all the 
        frames and loading them in memory if it is going to fail.
        """
        video = VideoParser.to_moviepy(video)
        extraction_type = FrameExtractionType.to_enum(extraction_type)
        
        if not NumberValidator.is_positive_number(frame_number_or_time):
            raise Exception('The provided "frame_number_or_time" is not a valid number.')
        
        if extraction_type == FrameExtractionType.FRAME_NUMBER:
            max_frame_number = video.fps * video.duration
            if frame_number_or_time > max_frame_number:
                raise Exception(f'The provided "frame_number" parameter {str(frame_number_or_time)}" is not valid. The maximum is {str(max_frame_number)}.')
            
            frame_number_or_time = frame_number_or_time * 1.0 / video.fps
        elif extraction_type == FrameExtractionType.TIME:
            if frame_number_or_time > video.duration:
                raise Exception(f'The provided "time" parameter {str(frame_number_or_time)} is not valid. The maximum is {str(video.duration)}.')
            
        return frame_number_or_time

    # TODO: I think I won't use this because I need to validate them first
    @classmethod
    def _get_frame(cls, video, frame_number_or_time, extraction_type: FrameExtractionType, output_filename: str = None):
        """
        Returns the frame of the provided 'video' according to the provided
        'frame_number_or_time' parameter and the desired 'extraction_type'.
        """
        video = VideoParser.to_moviepy(video)

        frame = video.get_frame(t = cls._get_frame_time(video, frame_number_or_time, extraction_type))

        if output_filename:
            cls._save_frame(frame, output_filename)

        return frame
    
    @classmethod
    def _get_frames(cls, video, frame_numbers_or_times, extraction_type: FrameExtractionType):
        """
        This method validates if all the provided 'frame_numbers_or_times'
        array parameter items are valid or not and them get all the frames
        and returns them as an array. This is optimum as it doesn't load
        frames until it's been verified the condition that all are valid.
        """
        video = VideoParser.to_moviepy(video)

        # Validate and get all times to obtain the frames later
        times = [cls._get_frame_time(video, frame_number_or_time, extraction_type) for frame_number_or_time in frame_numbers_or_times]

        return [video.get_frame(t = time) for time in times]
    
    @classmethod
    def get_frames_by_time(cls, video, times: list[float]):
        """
        Returns the frames (as numpy arrays) from the provided 'video'
        that corresponds to the provided 'times' time momnets of the
        video.

        This method will raise an Exception if the limits are not valid
        for the given 'video'.
        """
        return cls._get_frames(video, times, FrameExtractionType.TIME)

    @classmethod
    def _save_frame(cls, frame, output_filename: str):
        if not output_filename:
            return None
        
        # TODO: Validate 'output_filename'
        
        # TODO: This is how 'video.save_frame' works
        # https://github.com/Zulko/moviepy/blob/master/moviepy/video/VideoClip.py#L166
        # from imageio import imsave
        # im = self.get_frame(t)
        # if with_mask and self.mask is not None:
        #     mask = 255 * self.mask.get_frame(t)
        #     im = np.dstack([im, mask]).astype("uint8")
        # else:
        #     im = im.astype("uint8")

        imsave(output_filename, frame.astype("uint8"))

        return True

    @classmethod
    def get_frame_by_time(cls, video, time: float, output_filename: str = None):
        """
        Returns the frame (as numpy arrays) from the provided 'video'
        that corresponds to the provided 'time' moment.

        This method will raise an Exception if the limits are not valid
        for the given 'video'.
        """
        frame = cls.get_frames_by_time(video, [time])[0]

        if output_filename:
            cls._save_frame(frame, output_filename)

        return frame
    
    @classmethod
    def get_frames_by_time_from_start_to_end(cls, video, start_time: float, end_time: float):
        """
        Returns the frames (as numpy arrays) from the provided 'video'
        that corresponds to the lapse of time between the provided
        'start_time' and 'end_time'.

        This method will raise an Exception if the limits are not valid
        for the given 'video'.
        """
        video = VideoParser.to_moviepy(video)

        # Validate start and end parameters
        if not NumberValidator.is_positive_number(start_time) or start_time < 0 or start_time > video.duration:
            raise Exception(f'The provided "start_time" parameter "{str(start_time)}" is not valid. It must be a positive number between 0 and the video duration ({str(video.duration)}).')
        if not NumberValidator.is_positive_number(end_time) or end_time < 0 or end_time < start_time or end_time > video.duration:
            raise Exception(f'The provided "end_time" parameter "{str(end_time)}" is not valid. It must be a positive number between 0 and the video duration ({str(video.duration)}), and greater than the provided "start_time" parameter.')
        
        total_frames = int(video.duration * video.fps)
        time_moments = []
        for frame in range(total_frames):
            frame_time_moment = frame / video.fps  # tiempo en segundos para el frame actual
            if start_time <= frame_time_moment <= end_time:
                time_moments.append(frame_time_moment)

        return cls.get_frames_by_time(video, time_moments)

    @classmethod
    def get_frame_by_frame_number(cls, video, frame_number: int, output_filename: str = None):
        """
        Returns the frame (as numpy array) for the provided 'video' that
        corresponds to the provided 'frame_number'.

        This method will raise an Exception if the provided 'frame_number'
        is not a valid frame number for the given 'video'.
        """
        frame = cls._get_frame(video, frame_number, FrameExtractionType.FRAME_NUMBER)

        if output_filename:
            cls._save_frame(frame, output_filename)

        return frame
    
    @classmethod
    def get_frames_by_frame_number(cls, video, frame_numbers: list[int]):
        """
        Returns the frames (as numpy arrays) from the provided 'video'
        that corresponds to the provided 'frame_numbers'.

        This method will raise an Exception if the limits are not valid
        for the given 'video'.
        """
        return cls._get_frames(video, frame_numbers, FrameExtractionType.FRAME_NUMBER)
    
    @classmethod
    def get_frames_by_frame_number_from_start_to_end(cls, video, start_frame: float, end_frame: float):
        """
        Returns the frames (as numpy arrays) from the provided 'video' 
        from the 'start_frame' to the 'end_frame'.

        This method will raise an Exception if the limits are not valid
        for the given 'video'.
        """
        video = VideoParser.to_moviepy(video)

        max_frame_number = int(video.fps * video.duration)
        # Validate start and end parameters
        if not NumberValidator.is_positive_number(start_frame) or start_frame < 0 or start_frame > max_frame_number:
            raise Exception(f'The provided "start_frame" parameter "{str(start_frame)}" is not valid. It must be a positive number between 0 and the maximum frame ({str(max_frame_number)}).')
        if not NumberValidator.is_positive_number(end_frame) or end_frame < 0 or end_frame < start_frame or end_frame > max_frame_number:
            raise Exception(f'The provided "end_frame" parameter "{str(end_frame)}" is not valid. It must be a positive number between 0 and the maximum frame ({str(max_frame_number)}), and greater than the provided "start_time" parameter.')
        
        frames = [frame for frame in range(start_frame, end_frame + 1)]
        time_moments = [cls._get_frame_time(video, frame, FrameExtractionType.FRAME_NUMBER) for frame in frames]

        return cls.get_frames_by_time(video, time_moments)
    
    @classmethod
    def get_all_frames(cls, video):
        """
        Returns all the frames from the provided 'video'.
        """
        video = VideoParser.to_moviepy(video)

        # TODO: This was previously written in files like this:
        # if output_folder:
        #     video.write_images_sequence(output_folder + 'frame%05d.png')

        last_frame = (int) (video.duration * video.fps)

        return cls.get_frames_by_frame_number_from_start_to_end(video, 0, last_frame)
    
    @staticmethod
    def get_first_frame(video):
        """
        Obtain the first frame of the provided 'video' as a ndim=3
        numpy array containing the clip part (no mask) as not
        normalized values (between 0 and 255).
        """
        video = VideoParser.to_moviepy(video)

        # TODO: Some of the methods are failing (the ones with time
        # I think) so, when fixed, refactor this to make it light
        # weighting
        return VideoFrameExtractor.get_all_frames(video)[0]
    
    @staticmethod
    def get_last_frame(video):
        """
        Obtain the last frame of the provided 'video' as a ndim=3
        numpy array containing the clip part (no mask) as not
        normalized values (between 0 and 255).
        """
        video = VideoParser.to_moviepy(video)

        # TODO: Some of the methods are failing (the ones with time
        # I think) so, when fixed, refactor this to make it light
        # weighting
        all_frames = VideoFrameExtractor.get_all_frames(video)

        return all_frames[len(all_frames) - 1]
    
    # TODO: Would be perfect to have some methods to turn frames into
    # RGBA denormalized (0, 255) or normalized (0, 1) easier because
    # it is needed to work with images and other libraries. Those 
    # methods would iterate over the values and notice if they are in
    # an specific range so they need to be change or even if they are
    # invalid values (not even in [0, 255] range because they are not
    # rgb or rgba colors but math calculations).
    # This is actually being done by the VideoMaskHandler
    @classmethod
    def get_frame_as_rgba_by_time(cls, video, time: int, do_normalize: bool = False, output_filename: str = None):
        """
        Gets the frame of the requested 'time' of the provided
        'video' as a normalized RGBA numpy array that is built
        by joining the rgb frame (from main clip) and the alpha
        (from .mask clip), useful to detect transparent regions.
        """
        video = VideoParser.to_moviepy(video, has_mask = True)

        # We first normalize the clips
        main_frame = cls.get_frame_by_time(video, time) / 255  # RGB numpy array normalized 3d <= r,g,b
        mask_frame = cls.get_frame_by_time(video.mask, time)[:, :, np.newaxis]  # Alpha numpy array normalized 1d <= alpha
        # Combine RGB of frame and A from mask to RGBA numpy array (it is normalized)
        frame_rgba = np.concatenate((main_frame, mask_frame), axis = 2) # 4d <= r,g,b,alpha

        if output_filename:
            # TODO: Check extension
            ImageConverter.numpy_image_to_pil(frame_rgba).save(output_filename)
            # TODO: Write numpy as file image
            # Video mask is written as 0 or 1 (1 is transparent)
            # but main frame is written as 0 to 255, and the
            # 'numpy_image_to_pil' is expecting from 0 to 1
            # (normalized) instead of from 0 to 255 so it won't
            # work

        if not do_normalize:
            frame_rgba *= 255

        return frame_rgba



class CoordinateCorner:
    """
    Class that encapsulates a coordinate (x, y) representing the
    upper left corner of a video in a movement effect 
    envinronment. This class is used as type to represent the 
    position in which we want to place a video, considering this 
    position as the upper left corner of the video.
    """
    position = (0, 0)
    """
    The (x, y) tuple containing the position coordinate.
    """

    def __init__(self, x: int, y: int):
        if not x and x != 0:
            raise Exception('No "x" provided.')
        
        if not y and y != 0:
            raise Exception('No "y" provided.')
        
        # TODO: Check 'x' and 'y' are numbers and cast to int
        
        self.position = (x, y)

    def get_x(self):
        return self.position[0]
    
    def get_y(self):
        return self.position[1]
    
    def recalculate_for_video(self, background_video):
        """
        This method will return the coords (x, y) in which we need to place the
        a to have its upper left corner in the desired ('x', 'y') position over 
        the provided 'background_video' by making some calculations as below.

        Imagine a scene of a 1920x1080 black background and that the 'x' and 'y'
        you give as parameters are the upper left corner of the video. We will
        calculate to place that video there in the real situation, 
        over the 'background_video' that could be not 1920x1080.

        TODO: Maybe we could return an instance of a CoordinateCorner class, as
        it is what it is.
        """
        # TODO: Implement checkings

        # Considering a 1920x1080 scene, recalculate actual coords
        x = (int) (background_video.w * self.position[0] / 1920)
        y = (int) (background_video.h * self.position[1] / 1080)

        return (x, y)


from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionSetPosition, TFunctionResize, TFunctionRotate
from yta_multimedia.video.edition.effect.moviepy.position.objects.base_position_moviepy_effect import BasePositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.utils.position import get_moviepy_position
from yta_general_utils.math.rate_functions import RateFunction
from moviepy.editor import ColorClip
# TODO: Maybe unify all these to a MoviepyArgument (?) as
# they are very similar: 'start', 'end', 'rate_func'
# TODO: Maybe rename to MoviepyResize

class MoviepyResize:
    def __init__(self, initial_size: float, final_size: float, t_function: type = TFunctionResize.resize_from_to, rate_func: type = RateFunction.linear, *args, **kwargs):
        # TODO: Check that all params are provided and valid
        self.initial_size = initial_size
        self.final_size = final_size
        self.t_function = t_function
        self.rate_func = rate_func
        # TODO: Set '*args' and '**kwargs'

class MoviepySetPosition:
    def __init__(self, initial_position: tuple, final_position: float, t_function: type = TFunctionSetPosition.linear, rate_func: type = RateFunction.linear, *args, **kwargs):
        # TODO: Check that all params are provided and valid
        self.initial_position = initial_position
        self.final_position = final_position
        self.t_function = t_function
        self.rate_func = rate_func
        # TODO: Set '*args' and '**kwargs'

    def apply(self, video):
        BasePositionMoviepyEffect.validate_position(self.initial_position)
        BasePositionMoviepyEffect.validate_position(self.final_position)

        background_video = ColorClip(video.size, 'black', duration = video.duration)

        video = video.set_position(lambda t: self.t_function(t, video.duration, get_moviepy_position(video, background_video, self.initial_position), get_moviepy_position(video, background_video, self.final_position), self.rate_func))

        return video

class MoviepyRotate:
    # TODO: Check that all params are provided and valid
    def __init__(self, initial_rotation: int, final_rotation: int, t_function: type = TFunctionRotate.rotate_from_to, rate_func: type = RateFunction.linear, *args, **kwargs):
        self.initial_rotation = initial_rotation
        self.final_rotation = final_rotation
        self.t_function = t_function
        self.rate_func = rate_func
        # TODO: Set '*args' and '**kwargs'